// MultiTexture.cpp: implementation of the MultiTexture class.
//
//////////////////////////////////////////////////////////////////////

#include "MultiTexture.h"

PFNGLMULTITEXCOORD1FARBPROC		glMultiTexCoord1fARB	= NULL;
PFNGLMULTITEXCOORD2FARBPROC		glMultiTexCoord2fARB	= NULL;
PFNGLMULTITEXCOORD3FARBPROC		glMultiTexCoord3fARB	= NULL;
PFNGLMULTITEXCOORD4FARBPROC		glMultiTexCoord4fARB	= NULL;
PFNGLACTIVETEXTUREARBPROC		glActiveTextureARB		= NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC	glClientActiveTextureARB= NULL;	

int multitextureSupported = 0;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MultiTexture::MultiTexture()
{
	char *extensions;	
	extensions=strdup((char *) glGetString(GL_EXTENSIONS));			// Fetch Extension String
	int len=strlen(extensions);
	for (int i=0; i<len; i++)										// Separate It By Newline Instead Of Blank
		if (extensions[i]==' ') extensions[i]='\n';

	//MessageBox(hWnd,extensions,"supported GL extensions",MB_OK | MB_ICONINFORMATION);


	if( isInString(extensions,"GL_SGIS_texture_edge_clamp") ||
		isInString(extensions,"GL_SGI_texture_edge_clamp") ||
		isInString(extensions,"GL_EXT_texture_edge_clamp") )edge_clamp_supported=1;
	else edge_clamp_supported=0;

	if (  isInString(extensions,"GL_ARB_multitexture")				// Is Multitexturing Supported?
	&&    isInString(extensions,"GL_EXT_texture_env_add") )			// Is texture_env_add Supported?
//	&& isInString(extensions,"GL_EXT_texture_env_combine") )		// Is texture_env_combining Supported?
	// GL_EXT_texture_env_combine vzniklo z GL_EXT_texture_env_add

	{	
		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&maxTexelUnits);
		glMultiTexCoord1fARB	= (PFNGLMULTITEXCOORD1FARBPROC)		wglGetProcAddress("glMultiTexCoord1fARB");
		glMultiTexCoord2fARB	= (PFNGLMULTITEXCOORD2FARBPROC)		wglGetProcAddress("glMultiTexCoord2fARB");
		glMultiTexCoord3fARB	= (PFNGLMULTITEXCOORD3FARBPROC)		wglGetProcAddress("glMultiTexCoord3fARB");
		glMultiTexCoord4fARB	= (PFNGLMULTITEXCOORD4FARBPROC)		wglGetProcAddress("glMultiTexCoord4fARB");
		glActiveTextureARB		= (PFNGLACTIVETEXTUREARBPROC)		wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB= (PFNGLCLIENTACTIVETEXTUREARBPROC)	wglGetProcAddress("glClientActiveTextureARB");
		//MessageBox(hWnd,"The GL_ARB_multitexture extension will be used.","feature supported!",MB_OK | MB_ICONINFORMATION);
		multitextureSupported = 1;
		free(extensions);
		return ;//true;
	}
	else
	{
		maxTexelUnits=1;
		glMultiTexCoord1fARB	= NULL;
		glMultiTexCoord2fARB	= NULL;
		glMultiTexCoord3fARB	= NULL;
		glMultiTexCoord4fARB	= NULL;
		glActiveTextureARB		= NULL;
		glClientActiveTextureARB= NULL;
		multitextureSupported = 0;
		free(extensions);
	}
	//return false;
}

MultiTexture::~MultiTexture()
{

}

void MultiTexture::Write_extensions()
{
	char *extensions;
	int		len;
	FILE* sub;
	extensions=strdup((char *) glGetString(GL_EXTENSIONS));		// nacitanie rozsireni a skopirovanie do extensions
	len=strlen(extensions);
	for (int i=0; i<len; i++)if (extensions[i]==' ') extensions[i]='\n';	//nahradanie " " novym riadkom

	sub = fopen("rozsirenia.txt","w");
	fprintf(sub,"\nGL_VENDOR\n%s",(char *) glGetString(GL_VENDOR));
	fprintf(sub,"\nGL_RENDERER\n%s",(char *)glGetString(GL_RENDERER));
	fprintf(sub,"\nGL_VERSION\n%s",(char *) glGetString(GL_VERSION));
	//fprintf(sub,"\nGL_EXTENSIONS\n%s",(char *) glGetString(GL_EXTENSIONS));
	fprintf(sub,"\nGL_EXTENSIONS\n%s",extensions);
	fclose(sub);
	free(extensions);	
}

bool MultiTexture::isInString(char *string, const char *search) 
{
	int pos=0;
	int maxpos=strlen(search)-1;
	int len=strlen(string);	
	char *other;
	for (int i=0; i<len; i++) 
	{
		if ((i==0) || ((i>1) && string[i-1]=='\n')) 
		{							// New Extension Begins Here!
			other=&string[i];			
			pos=0;													// Begin New Search
			while (string[i]!='\n') 
			{														// Search Whole Extension-String
				if (string[i]==search[pos]) pos++;					// Next Position
				if ((pos>maxpos) && string[i+1]=='\n') return true; // We Have A Winner!
				i++;
			}			
		}
	}	
	return false;													// Sorry, Not Found!
}